@@ -1,4 +1,4 @@
-__version__ = "0.4.1"
+__version__ = "0.5.0"
 
 from .ssim import ssim, multi_scale_ssim, SSIMLoss, MultiScaleSSIMLoss
 from .msid import MSID
@@ -10,4 +10,7 @@ from .gs import GS
 from .isc import IS, inception_score
 from .vif import VIFLoss, vif_p
 from .brisque import BRISQUELoss, brisque
+from .perceptual import StyleLoss, ContentLoss, LPIPS, DISTS
 from .psnr import psnr
+from .fsim import fsim, FSIMLoss
+from .vsi import vsi, VSILoss
